/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert;

import org.rblasch.convert.converters.ChainedConverter;
import org.rblasch.convert.dijkstra.ShortestPathEngine;
import org.rblasch.convert.graph.Graphs;
import org.rblasch.convert.graph.Path;
import org.rblasch.convert.graph.WeightedDirectedGraph;
import org.rblasch.convert.type.Type;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

public class ShortestPathLookup implements ConverterLookup {
    private final WeightedDirectedGraph graph;

    public ShortestPathLookup(final WeightedDirectedGraph g) {
        this.graph = g;
    }

    public void addGraph(final WeightedDirectedGraph g) {
        Graphs.mergeInto(graph, g);
    }

    public void addType(final Type t) {
        graph.addVertex(new TypeVertex(t));
    }

    public void addConverter(final MetaConverter mc) {
        graph.addConnection(new TypeVertex(mc.getSourceType()),
                new ConverterEdge(mc, mc.getWeight()),
                new TypeVertex(mc.getDestinationType()));
    }

    public void addAllConverters(final Collection/*<MetaConverter>*/ mcs) {
        for (final Iterator i = mcs.iterator(); i.hasNext();) {
            addConverter((MetaConverter) i.next());
        }
    }

    public int getNumberOfKnownTypes() {
        return graph.getVertices().size();
    }

    public int getNumberOfKnownConversions() {
        return graph.getEdges().size();
    }

    public Set/*<Type>*/ getAllKnownTypes() {
        final Set/*<Type>*/ types = new HashSet();
        for (final Iterator i = graph.getVertices().iterator(); i.hasNext();) {
            final TypeVertex tv = (TypeVertex) i.next();
            types.add(tv.getType());
        }
        return types;
    }

    private List /*<MetaConverter>*/ getConvertersForPath(final Path path) {
        final List /*<MetaConverter>*/ converters = new LinkedList();

        for (final Iterator i = path.getEdges().iterator(); i.hasNext();) {
            final MetaConverter c = ((ConverterEdge) i.next()).getConverter();
            converters.add(c);
        }

        return converters;
    }

    private Path getShortestPath(final Type sourceType, final Type destinationType) throws NoSuchConversionException {
        final ShortestPathEngine spEngine = new ShortestPathEngine(graph);
        final Path shortestPath = spEngine.getShortestPath(new TypeVertex(sourceType), new TypeVertex(destinationType));
        if (shortestPath == null || shortestPath.getEdges().isEmpty()) {
            throw new NoSuchConversionException("No conversion from " + sourceType + " to " + destinationType + " in " + graph);
        }
        return shortestPath;
    }

    public MetaConverter lookup(final Type sourceType, final Type destinationType) throws NoSuchConversionException {
        return new ChainedConverter(getConvertersForPath(getShortestPath(sourceType, destinationType)));
    }

    public boolean equals(final Object obj) {
        return graph.equals(obj);
    }

    public int hashCode() {
        return graph.hashCode();
    }

    String dumpGraph() {
        return graph.toString();
    }
}
