/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert;

import org.rblasch.convert.type.Type;
import org.rblasch.convert.graph.SparseWeightedDirectedGraph;
import org.rblasch.convert.graph.WeightedDirectedGraph;
import org.rblasch.convert.converters.OnceConverterFactory;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

/**
 * @author Ronald Blaschke
 */
public class ShortestPathSearch implements ConverterLookup {
    private final Set/*<MetaConverter>*/ converters = new HashSet();
    private final Set/*<ConverterFactory>*/ factories = new HashSet();
    private final ShortestPathLookup lookup = new ShortestPathLookup(new SparseWeightedDirectedGraph());

    private static WeightedDirectedGraph convertersToGraph(final Set/*<MetaConverter>*/ cs) {
        final WeightedDirectedGraph g = new SparseWeightedDirectedGraph();
        for (final Iterator i = cs.iterator(); i.hasNext();) {
            final MetaConverter mc = (MetaConverter) i.next();
            g.addConnection(new TypeVertex(mc.getSourceType()),
                    new ConverterEdge(mc, mc.getWeight()),
                    new TypeVertex(mc.getDestinationType()));
        }
        return g;
    }

    private static void runFactories(final ShortestPathLookup lookup, final List/*<ConverterFactory>*/ factories, final Type sourceType, final Type destinationType) {
        for (final Iterator i = factories.iterator(); i.hasNext();) {
            final ConverterFactory f = (ConverterFactory) i.next();
            lookup.addGraph(convertersToGraph(f.create(sourceType, destinationType)));
        }
    }

    private static void runFactoriesOnAllKnownTypes(final ShortestPathLookup lookup, final Set/*<ConverterFactory>*/ factories) {
        final WeightedDirectedGraph g = new SparseWeightedDirectedGraph();
        final Set/*<Type>*/ allKnownTypes = lookup.getAllKnownTypes();
        for (final Iterator i = allKnownTypes.iterator(); i.hasNext();) {
            final Type sourceType = (Type) i.next();
            for (final Iterator j = allKnownTypes.iterator(); j.hasNext();) {
                final Type destinationType = (Type) j.next();
                for (final Iterator k = factories.iterator(); k.hasNext();) {
                    final ConverterFactory factory = (ConverterFactory) k.next();
                    for (final Iterator l = factory.create(sourceType, destinationType).iterator(); l.hasNext();) {
                        final MetaConverter mc = (MetaConverter) l.next();
                        g.addConnection(new TypeVertex(mc.getSourceType()),
                                new ConverterEdge(mc, mc.getWeight()),
                                new TypeVertex(mc.getDestinationType()));
                    }
                    //Graphs.mergeInto(g, convertersToGraph());
                }
            }
        }
        lookup.addGraph(g);
    }

    String dumpGraph() {
        return lookup.dumpGraph();
    }

    public void addConverter(final MetaConverter converter) {
        converters.add(converter);
        lookup.addConverter(converter);
    }

    public void addAllConverters(final Set/*<MetaConverter>*/ converters) {
        for (final Iterator i = converters.iterator(); i.hasNext();) {
            addConverter((MetaConverter) i.next());
        }
    }

    public void addOnceFactory(final OnceConverterFactory once) {
        lookup.addAllConverters(once.createOnce());
    }

    public void addFactory(final ConverterFactory factory) {
        if (factory instanceof OnceConverterFactory) {
            addOnceFactory((OnceConverterFactory) factory);
        }
        factory.setLookup(lookup);
        factories.add(factory);
    }

    public void addAllFactories(final Set/*<ConverterFactory>*/ factories) {
        for (final Iterator i = factories.iterator(); i.hasNext();) {
            addFactory((ConverterFactory) i.next());
        }
    }

    public MetaConverter lookup(final Type sourceType, final Type destinationType) {
        lookup.addType(sourceType);
        lookup.addType(destinationType);

        int numberOfTypes;
        NoSuchConversionException pendingEx;
        do {
            numberOfTypes = lookup.getNumberOfKnownTypes();
            try {
                return lookup.lookup(sourceType, destinationType);
            } catch (final NoSuchConversionException ex) {
                pendingEx = ex;
                runFactoriesOnAllKnownTypes(lookup, factories);
            }
        } while (numberOfTypes != lookup.getNumberOfKnownTypes());
        throw pendingEx;
    }
}
