/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert;

import org.rblasch.convert.converters.ObjectToMetaConvertersConverter;
import org.rblasch.convert.converters.OnceConverterFactory;
import org.rblasch.convert.converters.collections.ArrayConverterFactory;
import org.rblasch.convert.converters.collections.ArrayToListConverterFactory;
import org.rblasch.convert.converters.collections.ListConverterFactory;
import org.rblasch.convert.converters.lang.JlConverterFactory;
import org.rblasch.convert.converters.primitives.ByteToBooleanConverter;
import org.rblasch.convert.converters.primitives.StringToBooleanConverter;
import org.rblasch.convert.converters.primitives.StringToIntegerConverter;
import org.rblasch.convert.type.Type;
import org.rblasch.convert.type.Types;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * @author Ronald Blaschke
 * @api
 */
public class SptConverter implements TargetedConverter {
    private final ShortestPathSearch search = new ShortestPathSearch();

    private static Set/*<MetaConverter>*/ defaultConverters() {
        final Set/*<MetaConverter>*/ cs = new HashSet();
        // currently disabled, until basic stuff is working, to avoid converter cluttering
        if (false) {
            cs.add(new ByteToBooleanConverter());
            //cs.add(new IntegerToByteConverter());
            cs.add(new StringToIntegerConverter());
            cs.add(new StringToBooleanConverter());
        }
        return cs;
    }

    private static Set/*<ConverterFactory>*/ defaultFactories() {
        final Set/*<ConverterFactory>*/ cf = new HashSet();
        cf.add(new JlConverterFactory());
        cf.add(new ArrayToListConverterFactory());
        cf.add(new ArrayConverterFactory());
        cf.add(new ListConverterFactory());
        return cf;
    }

    public static SptConverter createEmpty() {
        return new SptConverter();
    }

    public static SptConverter createJl() {
        final SptConverter c = new SptConverter();
        c.addFactory(new JlConverterFactory());
        return c;
    }

    public static SptConverter createDefault() {
        return new SptConverter(defaultConverters(), defaultFactories());
    }

    public SptConverter() {
    }

    public SptConverter(final Set/*<MetaConverter>*/ converters) {
        search.addAllConverters(converters);
    }

    public SptConverter(final Set/*<MetaConverter>*/ converters, final Set/*<ConverterFactory>*/ factories) {
        search.addAllConverters(converters);
        search.addAllFactories(factories);
    }

    public void addConverter(final MetaConverter converter) {
        search.addConverter(converter);
    }

    public void addDynamicConverter(final Object converter) {
        final SptConverter c = new SptConverter();
        c.addConverter(new ObjectToMetaConvertersConverter(100));

        final Set/*<MetaConverter>*/ converters = (Set) c.convert(converter, Types.findPTypeByClass(Set.class, MetaConverter.class));
        for (final Iterator i = converters.iterator(); i.hasNext();) {
            final MetaConverter mc = (MetaConverter) i.next();
            search.addConverter(mc);
        }
    }

    public void addFactory(final ConverterFactory factory) {
        search.addFactory(factory);
    }

    public void addOnceFactory(final OnceConverterFactory once) {
        search.addOnceFactory(once);
    }

    public Set/*<MetaConverter>*/ removeAllConverters(final Type sourceType, final Type destinationType) {
        throw new UnsupportedOperationException("removeAllConverters not yet implemented");
    }

    public Object convert(final Object value, final Type destinationType) throws ConversionException {
        return convert(value, value == null ? Types.nullType() : Types.findTypeByClass(value.getClass()), destinationType);
    }

    public Object convert(final Object value, final Type sourceType, final Type destinationType) throws ConversionException {
        try {
            return search.lookup(sourceType, destinationType).getConverter()
                    .convert(value);
        } catch (final RuntimeException ex) {
            throw ex;  // propagate runtime exceptions
        } catch (final Exception ex) {
            throw new ConversionFailedException(ex);
        }
    }

    String dumpGraph() {
        return search.dumpGraph();
    }
}
