/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert.dijkstra;

import org.rblasch.convert.graph.Connection;
import org.rblasch.convert.graph.Edge;
import org.rblasch.convert.graph.ListPath;
import org.rblasch.convert.graph.Path;
import org.rblasch.convert.graph.Vertex;
import org.rblasch.convert.graph.WeightedDirectedGraph;
import org.rblasch.convert.graph.WeightedEdge;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * An implementation of Dijkstra's shortest path algorithm. It computes the shortest path (in distance)
 * to all cities in the map. The output of the algorithm is the shortest distance from the start vertex
 * to every other vertex, and the shortest path from the start vertex to every other.
 * <p/>
 * Upon calling
 * {@link #execute(Vertex, Vertex)},
 * the results of the algorithm are made available by calling
 * {@link #getPredecessor(Vertex)}
 * and
 * {@link #getShortestDistance(Vertex)}.
 * <p/>
 * To get the shortest path between the vertex <var>destination</var> and
 * the source vertex after running the algorithm, one would do:
 * <pre>
 * List l = new ArrayList();
 * <p/>
 * for (Object vertex = destination; vertex != null; vertex = engine.getPredecessor(vertex))
 * {
 *     l.add(vertex);
 * }
 * <p/>
 * Collections.reverse(l);
 * <p/>
 * return l;
 * </pre>
 *
 * @author Renaud Waldura &lt;renaud+tw@waldura.com&gt;
 * @version $Id: DijkstraEngine.java,v 1.4 2003/03/03 01:26:22 renaud Exp $
 * @see #execute(Vertex, Vertex)
 */
public class ShortestPathEngine {
    /**
     * Infinity value for distances.
     */
    public static final int INFINITE_DISTANCE = Integer.MAX_VALUE;

    /**
     * This comparator orders cities according to their shortest distances,
     * in ascending fashion. If two cities have the same shortest distance,
     * we compare the cities themselves.
     */
    private final Comparator shortestDistanceComparator = new Comparator() {
        public int compare(final Object left, final Object right) {
            // note that this trick doesn't work for huge distances, close to Integer.MAX_VALUE
            return getShortestDistance((Vertex) left) - getShortestDistance((Vertex) right);
        }
    };

    /**
     * The graph.
     */
    private final WeightedDirectedGraph graph;

    /**
     * The working set of cities, kept ordered by shortest distance.
     */
    private final Set /* of Vertex */ unsettledNodes = new HashSet();

    /**
     * The set of cities for which the shortest distance to the source
     * has been found.
     */
    private final Set /* of Vertex */ settledNodes = new HashSet();

    /**
     * The currently known shortest distance for all vertices.
     */
    private final Map /* of Vertex -> Integer */ shortestDistances = new HashMap();

    /**
     * Predecessors list: maps a vertex to its predecessor in the spanning tree of
     * shortest paths.
     */
    private final Map /* of Vertex -> Vertex */ predecessors = new HashMap();

    /**
     * Constructor.
     */
    public ShortestPathEngine(final WeightedDirectedGraph map) {
        this.graph = map;
    }

    /**
     * Initialize all data structures used by the algorithm.
     *
     * @param start the source node
     */
    private void init(final Vertex start) {
        settledNodes.clear();
        unsettledNodes.clear();

        shortestDistances.clear();
        predecessors.clear();

        // add source
        setShortestDistance(start, 0);
        unsettledNodes.add(start);
    }

    /**
     * Run Dijkstra's shortest path algorithm on the map.
     * The results of the algorithm are available through
     * {@link #getPredecessor(Vertex)}
     * and
     * {@link #getShortestDistance(Vertex)}
     * upon completion of this method.
     *
     * @param start       the starting vertex
     * @param destination the destination vertex. If this argument is <code>null</code>, the algorithm is
     *                    run on the entire graph, instead of being stopped as soon as the destination is reached.
     */
    public void execute(final Vertex start, final Vertex destination) {
        init(start);

        // the current node
        Vertex u;

        // extract the node with the shortest distance
        while ((u = extractMin()) != null) {
            //assert !isSettled(u);

            // destination reached, stop
            if (u == destination) break;

            markSettled(u);

            relaxNeighbors(u);
        }
    }

    public Path getShortestPath(final Vertex start, final Vertex end) {
        execute(start, end);

        final List /* of Vertex, Edge */ path = new LinkedList();
        if (getShortestDistance(end) == ShortestPathEngine.INFINITE_DISTANCE) {
            return null;
        }
        Vertex nextVertex = end;
        path.add(nextVertex);
        while (nextVertex != null) {
            final Vertex vertex = getPredecessor(nextVertex);
            if (vertex != null) {
                path.add(getMinWeightEdge(vertex, nextVertex));
                path.add(vertex);
            }
            nextVertex = vertex;
        }

        Collections.reverse(path);

        return new ListPath(path);
    }

    private Edge getMinWeightEdge(final Vertex start, final Vertex end) {
        return ((WeightedEdge) ((Connection) Collections.min(graph.getConnections(start, end), new Comparator() {
            public int compare(final Object o1, final Object o2) {
                final Connection c1 = (Connection) o1;
                final Connection c2 = (Connection) o2;

                return new Integer(((WeightedEdge) c1.getEdge()).getWeight()).compareTo(new Integer(((WeightedEdge) c2.getEdge()).getWeight()));
            }
        })).getEdge());
    }

    /**
     * Get the value of a segment.
     */
    public int getMinWeight(final Vertex start, final Vertex end) {
        return ((WeightedEdge) getMinWeightEdge(start, end)).getWeight();
    }

    /**
     * Extract the vertex with the currently shortest distance, and remove it from
     * the priority queue.
     *
     * @return the minimum vertex, or <code>null</code> if the queue is empty.
     */
    private Vertex extractMin() {
        if (unsettledNodes.isEmpty()) return null;

        final Object min = Collections.min(unsettledNodes, shortestDistanceComparator);
        unsettledNodes.remove(min);

        return (Vertex) min;
    }

    /**
     * Compute new shortest distance for neighboring nodes and update if a better
     * distance is found.
     *
     * @param u the node
     */
    private void relaxNeighbors(final Vertex u) {
        for (Iterator i = graph.getOutbound(u).iterator(); i.hasNext();) {
            final Vertex v = (Vertex) i.next();

            // skip node already settled
            if (isSettled(v)) continue;

            if (getShortestDistance(v) > getShortestDistance(u) + getMinWeight(u, v)) {
                // assign new shortest distance and mark unsettled
                setShortestDistance(v, getShortestDistance(u) + getMinWeight(u, v));

                // assign predecessor in shortest path
                setPredecessor(v, u);
            }
        }
    }

    /**
     * Mark a node as settled.
     *
     * @param u the node
     */
    private void markSettled(final Vertex u) {
        settledNodes.add(u);
    }

    /**
     * Test a node.
     *
     * @param v the node to consider
     * @return whether the node is settled, ie. its shortest distance
     *         has been found.
     */
    private boolean isSettled(final Vertex v) {
        return settledNodes.contains(v);
    }

    /**
     * @return the shortest distance from the source to the given vertex, or
     *         {@link ShortestPathEngine#INFINITE_DISTANCE} if there is no route to the destination.
     */
    public int getShortestDistance(final Vertex vertex) {
        final Number d = (Number) shortestDistances.get(vertex);
        return (d == null) ? INFINITE_DISTANCE : d.intValue();
    }

    /**
     * Set the new shortest distance for the given node,
     * and re-balance the set according to new shortest distances.
     *
     * @param vertex   the node to set
     * @param distance new shortest distance value
     */
    private void setShortestDistance(final Vertex vertex, final int distance) {
        // this crucial step ensure no duplicates will be created in the queue
        // when an existing unsettled node is updated with a new shortest distance
        unsettledNodes.remove(vertex);

        shortestDistances.put(vertex, new Integer(distance));

        // re-balance the sorted set according to the new shortest distance found
        // (see the comparator the set was initialized with)
        unsettledNodes.add(vertex);
    }

    /**
     * @return the vertex leading to the given vertex on the shortest path, or
     *         <code>null</code> if there is no route to the destination.
     */
    public Vertex getPredecessor(final Vertex vertex) {
        return (Vertex) predecessors.get(vertex);
    }

    private void setPredecessor(final Vertex a, final Vertex b) {
        predecessors.put(a, b);
    }
}
