/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert.graph;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

public final class SparseWeightedDirectedGraph implements WeightedDirectedGraph {
    private final Set /* of Vertex */ vertices = new HashSet();
    private final Set /* of WeightedDirectedEdge */ edges = new HashSet();
    private final Set /* of Connection */ connections = new HashSet();

    public Set /* of Vertex */ getVertices() {
        return Collections.unmodifiableSet(vertices);
    }

    public Set /* of WeightedDirectedEdge */ getEdges() {
        return Collections.unmodifiableSet(edges);
    }

    public void addVertex(final Vertex v) {
        vertices.add(v);
    }

    /**
     * Enter a new segment in the graph.
     */
    public void addConnection(final Vertex start, final Edge edge, final Vertex end) {
        vertices.add(start);
        edges.add(edge);
        vertices.add(end);
        connections.add(new Connection(start, edge, end));
    }

    public Set/*<Connection>*/ getAllConnections() {
        return Collections.unmodifiableSet(connections);
    }

    public Set /* of Connection */ getConnections(final Vertex start, final Vertex end) {
        final Set /* of Connection */ cons = new HashSet();
        for (final Iterator i = connections.iterator(); i.hasNext();) {
            final Connection c = (Connection) i.next();
            if (c.getStart().equals(start) && c.getEnd().equals(end)) {
                cons.add(c);
            }
        }
        return cons;
    }

    /**
     * Get the list of vertices that can be reached from the given vertex.
     */
    public Set /* Vertex */ getOutbound(final Vertex vertex) {
        final Set destinations = new HashSet();

        for (final Iterator i = connections.iterator(); i.hasNext();) {
            final Connection c = (Connection) i.next();
            if (c.getStart().equals(vertex)) {
                destinations.add(c.getEnd());
            }
        }

        return destinations;
    }

    /**
     * Get the list of vertices that lead to the given vertex.
     */
    public Set /* of Vertex */ getInbound(final Vertex vertex) {
        final Set predecessors = new HashSet();

        for (final Iterator i = connections.iterator(); i.hasNext();) {
            final Connection c = (Connection) i.next();
            if (c.getEnd().equals(vertex)) {
                predecessors.add(c.getStart());
            }
        }

        return predecessors;
    }

    public boolean equals(final Object obj) {
        if (this == obj) return true;
        if (!(obj instanceof SparseWeightedDirectedGraph)) return false;

        final SparseWeightedDirectedGraph rhs = (SparseWeightedDirectedGraph) obj;
        if (!edges.equals(rhs.edges)) return false;
        if (!vertices.equals(rhs.vertices)) return false;

        return true;
    }

    public int hashCode() {
        int result = vertices.hashCode();
        result = 29 * result + edges.hashCode();
        return result;
    }

    public String toString() {
        return connections.toString();
    }
}
