/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert.type;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

/**
 * @author Ronald Blaschke
 */
public class ParameterizableType extends AbstractType {
    static ParameterizableType valueOf(final String s) throws UnknownTypeException {
        if (s.endsWith("[]")) {
            // special array syntax, eg int[], java.lang.Integer[]
            final String componentName = s.substring(0, s.length() - 2);
            final Type componentType = Types.findTypeByName(componentName);

            return new ParameterizableType(Types.findTypeByName("array"), componentType);
        } else {
            final int ltPos = s.indexOf('<');
            final int gtPos = s.lastIndexOf('>');

            if (ltPos == -1 || gtPos == -1) {
                throw new UnknownTypeException(s + " is not a parameterizable type");
            }

            final String parameters = s.substring(ltPos + 1, gtPos);

            final Type mainType = Types.findTypeByName(s.substring(0, ltPos));
            final List/*<Type>*/ parameterTypes = new ArrayList();
            for (final StringTokenizer tokenizer = new StringTokenizer(parameters, ","); tokenizer.hasMoreElements();) {
                final String token = tokenizer.nextToken();
                parameterTypes.add(Types.findTypeByName(token.trim()));
            }

            return new ParameterizableType(mainType, parameterTypes);
        }
    }

    private final Type t;
    private final List/*<Type>*/ parameters;

    ParameterizableType(final Type t, final Type p1) {
        this(t, Collections.singletonList(p1));
    }

    ParameterizableType(final Type t, final Type p1, final Type p2) {
        this(t, Arrays.asList(new Type[]{p1, p2}));
    }

    ParameterizableType(final Type t, final List/*<Type>*/ parameters) {
        this.t = t;
        this.parameters = new ArrayList(parameters);
    }

    public Type getType() {
        return t;
    }

    public List/*<Type>*/ getParameters() {
        return parameters;
    }

    public Type getParameter(final int i) {
        return (Type) parameters.get(i);
    }

    public List/*<Type>*/ toList() {
        final List/*<Type>*/ list = new ArrayList(1 + parameters.size());
        list.add(t);
        list.addAll(parameters);
        return list;
    }

    public String getName() {
        final StringBuffer buffy = new StringBuffer();
        buffy.append(t.getName());
        buffy.append('<');
        for (final Iterator i = parameters.iterator(); i.hasNext();) {
            buffy.append(((Type) i.next()).getName());
            if (i.hasNext()) {
                buffy.append(",");
            }
        }
        buffy.append('>');
        return buffy.toString();
    }
}
