/*
 * Copyright 2004 Ronald Blaschke.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.rblasch.convert;

import junit.framework.TestCase;
import org.rblasch.convert.converters.ClassMetaConverter;
import org.rblasch.convert.converters.lang.StringToClassConverter;
import org.rblasch.convert.converters.lang.NullToStringConverter;
import org.rblasch.convert.converters.primitives.ByteToBooleanConverter;
import org.rblasch.convert.converters.primitives.IntegerToByteConverter;
import org.rblasch.convert.converters.primitives.StringToBooleanConverter;
import org.rblasch.convert.converters.primitives.StringToIntegerConverter;
import org.rblasch.convert.type.Types;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * @author Ronald Blaschke
 */
public class ConverterTest extends TestCase {
    public void testNoIdentity() throws Exception {
        final SptConverter c = SptConverter.createEmpty();
        try {
            c.convert("ABC", Types.findTypeByClass(String.class));
            fail("Should throw no such conversion exception");
        } catch (final NoSuchConversionException ex) {
            // expected
        }
    }

    public void testNull() throws Exception {
        final SptConverter c = SptConverter.createEmpty();
        c.addConverter(new NullToStringConverter());
        assertEquals("null", c.convert(null, Types.findTypeByClass(String.class)));
    }

    public void testSimple() throws Exception {
        final Set/*<MetaConverter>*/ converters = new HashSet();
        converters.add(new ClassMetaConverter(Byte.class, Integer.class,
                new Converter() {
                    public Object convert(final Object obj) throws Exception {
                        return new Integer(((Byte) obj).byteValue());
                    }
                }, 10));

        final SptConverter ci = new SptConverter(converters);

        final Object i = ci.convert(new Byte((byte) 42), Types.findTypeByClass(Integer.class));
        assertEquals(Integer.class, i.getClass());
        assertEquals(42, ((Integer) i).intValue());
    }

    public void testToString() throws Exception {
        final SptConverter ci = SptConverter.createJl();

        final Object s = ci.convert(new Integer(42), Types.findTypeByClass(String.class));
        assertEquals(String.class, s.getClass());
        assertEquals("42", s);
    }

    public void testInterface() throws Exception {
        final SptConverter ci = SptConverter.createJl();
        final LinkedList list = new LinkedList();
        list.add(new Integer(42));

        final Object l = ci.convert(list, Types.findTypeByClass(List.class));
        assertEquals(LinkedList.class, l.getClass());
        assertEquals(new Integer(42), ((List) l).get(0));
    }

    public void testStringToInteger() throws Exception {
        final SptConverter ci = new SptConverter(Collections.singleton(new StringToIntegerConverter()));
        final String s = "42";

        final Object i = ci.convert(s, Types.findTypeByClass(Integer.class));
        assertEquals(new Integer(42), i);
    }

    public void testStringToBoolean() throws Exception {
        final Set/*<MetaConverter>*/ cs = new HashSet();
        cs.add(new ByteToBooleanConverter());
        cs.add(new IntegerToByteConverter());
        cs.add(new StringToIntegerConverter());

        final SptConverter ci = new SptConverter(cs);
        assertEquals(Boolean.TRUE, (Boolean) ci.convert("1", Types.findTypeByClass(Boolean.class)));
        assertEquals(Boolean.FALSE, (Boolean) ci.convert("0", Types.findTypeByClass(Boolean.class)));
    }

    public void testStringToBoolean2() throws Exception {
        final Set/*<MetaConverter>*/ cs = new HashSet();
        // possible conversion: String -> Integer -> Byte -> Boolean
        cs.add(new ByteToBooleanConverter());
        cs.add(new IntegerToByteConverter());
        cs.add(new StringToIntegerConverter());

        // but this one is shorter, and will be used
        cs.add(new StringToBooleanConverter());

        final SptConverter ci = new SptConverter(cs);
        assertEquals(Boolean.TRUE, (Boolean) ci.convert("1", Types.findTypeByClass(Boolean.class)));
        assertEquals(Boolean.FALSE, (Boolean) ci.convert("0", Types.findTypeByClass(Boolean.class)));
    }

    public void testWidening() throws Exception {
        final SptConverter ci = SptConverter.createJl();
        assertEquals(new Long(10), ci.convert(new Byte((byte) 10), Types.findTypeByClass(Long.class)));
    }

    public void testAddConverter() throws Exception {
        final SptConverter ci = new SptConverter();
    }

    public void testClass() {
        final Set/*<MetaConverter>*/ cs = new HashSet();
        cs.add(new StringToClassConverter());

        final SptConverter ci = new SptConverter(cs);
        final Object o = ci.convert(A.class.getName(), Types.findTypeByClass(Class.class));
        assertTrue(o instanceof Class);
        assertEquals(A.class.getName(), ((Class) o).getName());
    }

    private static class A {}
}
